/* This is a simple test client program to request a stock quote from   */
/* the stock quote server application (server.c).  The quote request    */
/* is sent to the server using HTTP / SOAP / XML.  The client then      */
/* waits for a response from the server, displays the result, and       */
/* exits..                                                              */

#include <stdio.h>
#include <stdlib.h>
#include "rtxsrc/rtxSOAP.h"
#include "rtxsrc/rtxSocket.h"
#include "rtxsrc/rtxStreamMemory.h"
#include "rtxsrc/rtxStreamSocket.h"
#include "StockQuoteWSDL.h"

int main (int argc, char** argv)
{
   GetLastTradePrice_Input req;
   GetLastTradePrice_Output resp;
   GetLastTradePrice_Fault fault;

   OSCTXT	ctxt;
   OSSOAPCONN   soapConn;
   OSOCTET*     ptmpbuf = 0;
   int		i, stat;
   const char*  url = "http://127.0.0.1:9123";
   OSBOOL       verbose = FALSE, trace = TRUE;

   /* Process command line arguments */

   if (argc > 1) {
      for (i = 1; i < argc; i++) {
         if (!strcmp (argv[i], "-v")) verbose = TRUE;
         else if (!strcmp (argv[i], "-url")) url = argv[++i];
         else if (!strcmp (argv[i], "-notrace")) trace = FALSE;
         else {
            printf ("usage: client [-v] [-url <url>] [-notrace]\n");
            printf ("   -v    verbose mode: print trace info\n");
            printf ("   -url  connect to <url>\n");
            printf ("   -notrace  do not display trace info\n");
            return 0;
         }
      }
   }

   /* Init context */

   stat = rtXmlInitContext (&ctxt);
   if (0 != stat) {
      printf ("Context initialization failed.\n");
      rtxErrPrint (&ctxt);
      return stat;
   }

   if (verbose) 
      rtxSetDiag (&ctxt, 1);

   /* One-time socket package initialization (required for Windows only) */

   stat = rtxSocketsInit();
   if (0 != stat) {
      printf ("Socket initialization failed.\n");
      rtxErrSetData (&ctxt, stat, __FILE__, __LINE__);
      rtxErrPrint (&ctxt);
      return stat;
   }

   /* Populate structure of generated type */

   Init_GetLastTradePrice_Input (&ctxt, &req);
   req.TradePriceRequest.tickerSymbol.value = (const OSUTF8CHAR*)"DIS";

   /* Encode */

   stat = rtXmlSetEncBufPtr (&ctxt, 0, 0);

   if (0 == stat)
      stat = XmlE_GetLastTradePrice_Input (&ctxt, &req);

   if (0 != stat) {
      printf ("Encoding failed\n");
      rtxErrPrint (&ctxt);
      return stat;
   }

   /* Connect to StockQuote server */

   rtxSoapInitConn (&soapConn, &ctxt, OSSOAPV1, url);

   stat = rtxSoapConnect (&soapConn);
   if (stat != 0) {
      rtxErrPrint (&ctxt);
      return stat;
   }
   if (trace) printf ("Connection to web service opened\n");

   /* Send request */

   stat = rtxSoapSendHttp (&soapConn, rtXmlGetEncBufPtr(&ctxt));
   if (stat != 0) {
      rtxErrPrint (&ctxt);
      return stat;
   }
   if (trace) printf ("HTTP request sent\n");

   /* Wait for response */

   stat = rtxSoapRecvHttp (&soapConn);
   if (stat != 0) {
      rtxErrPrint (&ctxt);
      rtxStreamClose (&ctxt);
      return stat;
   }
   if (trace) printf ("HTTP response received\n");

   /* Read HTTP response into memory */

   stat = rtxSoapRecvHttpContent (&soapConn, &ptmpbuf);
   if (stat != 0) {
      rtxErrPrint (&ctxt);
      return stat;
   }

   /* Close connection */

   rtxStreamClose (&ctxt);

   /* Decode SOAP XML message */

   stat = rtxStreamMemoryAttach 
      (&ctxt, ptmpbuf, soapConn.contentLength, OSRTSTRMF_INPUT);

   if (0 != stat) {
      printf ("rtxStreamMemoryAttach failed\n");
      rtxErrPrint (&ctxt);
      return stat;
   }

   stat = XmlD_GetLastTradePrice_Output (&ctxt, &resp, &fault);

   if (0 == stat) {
      if (trace) {
         printf ("Decode was successful\n");
         printf ("Decoded record:\n");
         Print_GetLastTradePrice_Output 
            ("GetLastTradePriceOutput", &resp);
      }
   }
   else if (trace) {
      printf ("decode failed\n");
      rtxErrPrint (&ctxt);
   }

   rtxFreeContext (&ctxt);

   return stat;
}
