/*
 * Copyright (c) 2003-2018 Objective Systems, Inc.
 *
 * This software is furnished under a license and may be used and copied
 * only in accordance with the terms of such license and with the
 * inclusion of the above copyright notice. This software or any other
 * copies thereof may not be provided or otherwise made available to any
 * other person. No title to and ownership of the software is hereby
 * transferred.
 *
 * The information in this software is subject to change without notice
 * and should not be construed as a commitment by Objective Systems, Inc.
 *
 * PROPRIETARY NOTICE
 *
 * This software is an unpublished work subject to a confidentiality agreement
 * and is protected by copyright and trade secret law.  Unauthorized copying,
 * redistribution or other use of this work is prohibited.
 *
 * The above notice of copyright on this source code product does not indicate
 * any actual or intended publication of such source code.
 *
 *****************************************************************************/
/*! \page dompage C DOM Runtime Library Functions
 * The <b>C run-time DOM library</b> contains functions used to encode/decode
 * XML data in a DOM tree.  These functions are identified by their
 * <i>rtDom</i> prefixes.
 *
 * The categories of functions provided are as follows:<ul>
 * <li>Functions to add nodes.</li>
 * <li>Functions to add attributes to a node.</li>
 * <li>Functions to add namespace and xsi:type information.</li>
 * <li>Functions to encode data types to DOM.</li>
 * </ul>
 */
/**
 * @file osrtdom.h
 * DOM low-level C encode/decode functions.
 */
#ifndef _OSRTDOM_H_
#define _OSRTDOM_H_

#include "rtxsrc/rtxCommon.h"
#include "rtdomsrc/rtxDomDefs.h"
#include "rtxmlsrc/osrtxml.h"
#include "rtdomsrc/domAPI.h"
#include "rtxmlsrc/rtXmlNamespace.h"

/**
 * @defgroup rtDom DOM runtime encode/decode functions.
 * @{
 */

/* This is necessary to disable a Visual C++ warning about an unreferenced
   formal parameter passed to the rtDom* functions. */
#ifdef _MSC_VER
#pragma warning(disable: 4100)
#endif /* _MSC_VER */

#ifdef __cplusplus
extern "C" {
#endif

/**
 * Adds attribute to the node. The value is taken from the context's buffer
 * (pctxt->buffer.data with the length pctxt->buffer.byteIndex).
 * This function resets pctxt->buffer.byteIndex to 0.
 *
 * @param pctxt        A pointer to a context structure. This provides a
 *                       storage area for the function to store all working
 *                       variables that must be maintained between function
 *                       calls.
 * @param domDoc       A pointer to a DOM-document.
 * @param node         A pointer to a node where attribute to be added.
 * @param attrName     A pointer to attribute's name.
 * @param pNS          XML namespace information (prefix and URI).
 * @param pNSAttrs     List of namespace attributes to be added to element.
 * @return             Completion status of operation:
 *                       - 0 = success,
 *                       - negative return value is error.
 */
EXTERNDOM int rtDomAddAttr (OSCTXT* pctxt,
                            OSRTDOMDocPtr domDoc,
                            OSRTDOMNodePtr node,
                            const OSUTF8CHAR* attrName,
                            OSXMLNamespace* pNS,
                            OSRTDList* pNSAttrs);

/**
 * Adds child node to the parent's node. The content (text) is taken from the
 * context's buffer (pctxt->buffer.data with the length pctxt->buffer.byteIndex).
 * If pctxt->buffer.byteIndex is 0 then empty element is added.
 * This function resets pctxt->buffer.byteIndex to 0.
 *
 * @param pctxt        A pointer to a context structure. This provides a
 *                       storage area for the function to store all working
 *                       variables that must be maintained between function
 *                       calls.
 * @param domDoc       A pointer to a DOM-document.
 * @param parentNode   A pointer to a parent node.
 * @param elemName     A pointer to element's name.
 * @param pNS          XML namespace information (prefix and URI).
 * @param pNSAttrs     List of namespace attributes to be added to element.
 * @return             Completion status of operation:
 *                       - 0 = success,
 *                       - negative return value is error.
 */
EXTERNDOM int rtDomAddNode (OSCTXT* pctxt,
                            OSRTDOMDocPtr domDoc,
                            OSRTDOMNodePtr parentNode,
                            const OSUTF8CHAR* elemName,
                            OSXMLNamespace* pNS,
                            OSRTDList* pNSAttrs);

/**
 * Adds namespace attributes to the root node.
 *
 * @param pctxt        A pointer to a context structure. This provides a
 *                       storage area for the function to store all working
 *                       variables that must be maintained between function
 *                       calls.
 * @param domDoc       A pointer to a DOM-document.
 * @param rootNode     A pointer to a root node.
 * @param pNSAttrs     A pointer to a list of namespace attrs.
 * @return             Completion status of operation:
 *                       - 0 = success,
 *                       - negative return value is error.
 */
EXTERNDOM int rtDomAddNSAttrs
   (OSCTXT* pctxt, OSRTDOMDocPtr domDoc, OSRTDOMNodePtr rootNode,
    OSRTDList* pNSAttrs);

/**
 * Adds XSI attributes to the node.
 *
 * @param pctxt        A pointer to a context structure. This provides a
 *                       storage area for the function to store all working
 *                       variables that must be maintained between function
 *                       calls.
 * @param domDoc       A pointer to a DOM-document.
 * @param node         A pointer to a node.
 * @param needXSI      Determines whether XSI namespace declaration needed.
 * @return             Completion status of operation:
 *                       - 0 = success,
 *                       - negative return value is error.
 */
EXTERNDOM int rtDomEncXSIAttrs
   (OSCTXT* pctxt, OSRTDOMDocPtr domDoc, OSRTDOMNodePtr node,
    OSBOOL needXSI);

/**
 * This function starts decoding of the DOM document.
 *
 * @param domDoc       A pointer to a DOM-document.
 * @param pSaxBase     A pointer to SAX handler's base.
 * @return             Completion status of operation:
 *                       - 0 = success,
 *                       - negative return value is error.
 */
EXTERNDOM int rtDomDecodeDoc
   (OSRTDOMDocPtr domDoc, struct OSSAXHandlerBase* pSaxBase);

/**
 * This function encodes a variable of the XSD string type.
 *
 * @param pctxt        Pointer to context block structure.
 * @param pvalue       Null-terminated XML string value to be encoded.
 * @return             Completion status of operation:
 *                       - 0 = success,
 *                       - negative return value is error.
 */
EXTERNDOM int rtDomEncStringValue (OSCTXT* pctxt,
                                   const OSUTF8CHAR* pvalue);

/**
 * This function encodes a variable of the XSD string type.
 *
 * @param pctxt        Pointer to context block structure.
 * @param pvalue       XML string value to be encoded.
 * @return             Completion status of operation:
 *                       - 0 = success,
 *                       - negative return value is error.
 */
EXTERNDOM int rtDomEncString (OSCTXT* pctxt, OSXMLSTRING* pvalue);

/**
 * This function encodes a variable of the XSD any type.  This is considered
 * to be a fully-wrapped element of any type (for example:
 * \<myType\>myData\</myType\>)
 *
 * @param pctxt        Pointer to context block structure.
 * @param domDoc       A pointer to a DOM-document.
 * @param parentNode   A pointer to a parent node.
 * @param pXmlData     Value to be encoded. This is a string containing
 *                       the fully-encoded XML text to be copied to the
 *                       output stream.
 * @param elemName     XML element name.  A name must be provided.
 *                       If NULL pointer is passed, no element tag is
 *                       added to the encoded value.
 * @param pNS          XML namespace information (prefix and URI).
 * @param pNSAttrs     List of namespace attributes to be added to element.
 * @return             Completion status of operation:
 *                       - 0 = success,
 *                       - negative return value is error.
 */
EXTERNDOM int rtDomEncAny (OSCTXT* pctxt, OSRTDOMDocPtr domDoc,
                           OSRTDOMNodePtr parentNode,
                           const OSXMLSTRING* pXmlData,
                           const OSUTF8CHAR* elemName,
                           OSXMLNamespace* pNS,
                           OSRTDList* pNSAttrs);

/**
 * This function encodes a list of OSAnyAttr attributes in which
 * the name and value are given as a UTF-8 string.
 *
 * @param pctxt        Pointer to context block structure.
 * @param domDoc       A pointer to a DOM-document.
 * @param node         A pointer to a node where attributes to be added.
 * @param pAnyAttrList List of attributes.
 * @return             Completion status of operation:
 *                       - 0 = success,
 *                       - negative return value is error.
 */
EXTERNDOM int rtDomEncAnyAttr (OSCTXT* pctxt, OSRTDOMDocPtr domDoc,
                               OSRTDOMNodePtr node, OSRTDList* pAnyAttrList);

/**
 * Adds content or CDATA section to the node. The content (text) is taken from
 * the context's buffer (pctxt->buffer.data with the length
 * pctxt->buffer.byteIndex).
 * This function resets pctxt->buffer.byteIndex to 0.
 *
 * @param pctxt        A pointer to a context structure. This provides a
 *                       storage area for the function to store all working
 *                       variables that must be maintained between function
 *                       calls.
 * @param domDoc       A pointer to a DOM-document.
 * @param curNode      A pointer to a current node.
 * @return             Completion status of operation:
 *                       - 0 = success,
 *                       - negative return value is error.
 */
EXTERNDOM int rtDomSetNode (OSCTXT* pctxt,
                            OSRTDOMDocPtr domDoc,
                            OSRTDOMNodePtr curNode);

/**
 * This function adds the subtree to the specified node.
 * The pXmlData contains the fragment of XML data. This fragment
 * is parsed to sub-tree and then inserted as children to the 'node'.
 *
 * @param pctxt   Pointer to the context structure. This context might
 *                  be used for memory allocations, if necessary.
 * @param xmlDoc  Pointer to the document
 * @param node    Pointer to the node where sub-tree to be inserted.
 * @param pXmlData Pointer to UTF-8 string representing
 *                   fragment of XML data. For example, "<elem>data</elem>".
 * @param dataLen Length of XML data.
 * @return             Completion status of operation:
 *                       - 0 = success,
 *                       - negative return value is error.
 */
EXTERNDOM int rtDomAddSubTree (OSCTXT* pctxt, OSRTDOMDocPtr xmlDoc,
                                      OSRTDOMNodePtr node,
                                      const OSUTF8CHAR* pXmlData,
                                      size_t dataLen);


#ifdef __cplusplus
}
#endif
/**
 * @}
 */
#endif /* _OSRTDOM_H_ */

